/*
** $Id: agnxlib.h, 2009/03/01 15:41:13 $
** Auxiliary functions for building Agena libraries
** See Copyright Notice in agena.h
*/


#ifndef agnxlib_h
#define agnxlib_h


#include <stddef.h>
#include <stdio.h>

#include "agena.h"
#include "llimits.h"  /* for defines: lua_assert, lua_lock, lua_unlock */


#if defined(LUA_COMPAT_GETN)
LUALIB_API int (luaL_getn) (lua_State *L, int t);
LUALIB_API void (luaL_setn) (lua_State *L, int t, int n);
#else
#define luaL_getn(L,i)          ((int)lua_objlen(L, i))
#define luaL_setn(L,i,j)        ((void)0)  /* no op! */
#endif

#if defined(LUA_COMPAT_OPENLIB)
#define luaI_openlib	luaL_openlib
#endif


/* extra error code for `luaL_load' */
#define LUA_ERRFILE     (LUA_ERRERR+1)

typedef struct luaL_Reg {
  const char *name;
  lua_CFunction func;
} luaL_Reg;



LUALIB_API void (luaI_openlib) (lua_State *L, const char *libname,
                                const luaL_Reg *l, int nup);
LUALIB_API void (luaL_register) (lua_State *L, const char *libname,
                                const luaL_Reg *l);
LUALIB_API int (luaL_getmetafield) (lua_State *L, int obj, const char *e);
LUALIB_API int (luaL_callmeta) (lua_State *L, int obj, const char *e);
LUALIB_API int (luaL_typerror) (lua_State *L, int narg, const char *tname);
LUALIB_API int (luaL_argerror) (lua_State *L, int numarg, const char *extramsg);
LUALIB_API int (luaL_typeerror) (lua_State *L, int narg, const char *extramsg, int type);  /* Agena 1.1.0 */

LUALIB_API const char *(luaL_checklstring) (lua_State *L, int numArg,
                                                          size_t *l);
LUALIB_API const char *(luaL_optlstring) (lua_State *L, int numArg,
                                          const char *def, size_t *l);

LUALIB_API lua_Number (luaL_checknumber) (lua_State *L, int numArg);
LUALIB_API lua_Number (luaL_optnumber) (lua_State *L, int nArg, lua_Number def);
LUALIB_API int        (agnL_checkboolean) (lua_State *L, int narg);
LUALIB_API lua_Number (agnL_checknumber) (lua_State *L, int numArg);  /* Agena 1.4.3/1.5.0 */
LUALIB_API lua_Number (agnL_optnumber) (lua_State *L, int nArg, lua_Number def);  /* Agena 1.4.3/1.5.0 */
LUALIB_API int        (agnL_optboolean) (lua_State *L, int narg, int def);  /* Agena 1.6.0 */

LUALIB_API lua_Integer (luaL_checkinteger) (lua_State *L, int numArg);
LUALIB_API lua_Integer (luaL_optinteger) (lua_State *L, int nArg, lua_Integer def);
LUALIB_API lua_Integer (agnL_optposint) (lua_State *L, int nArg, lua_Integer def);  /* 2.9.4 */
LUALIB_API lua_Integer (agnL_optnonnegint) (lua_State *L, int narg, lua_Integer def);  /* 2.10.0 */
LUALIB_API lua_Integer (agnL_checkinteger) (lua_State *L, int numArg);  /* Agena 1.4.3/1.5.0 */
LUALIB_API lua_Integer (agnL_optinteger) (lua_State *L, int nArg,  /* Agena 1.4.3/1.5.0 */
                                          lua_Integer def);
LUALIB_API lua_Number (agnL_optpositive) (lua_State *L, int narg, lua_Number def);   /* Agena 2.12.5 */
LUALIB_API lua_Number (agnL_optnonnegative) (lua_State *L, int narg, lua_Number def);  /* Agena 2.12.5 */

LUALIB_API int32_t (luaL_optint32_t) (lua_State *L, int narg, int32_t def);
LUALIB_API off64_t (luaL_optoff64_t) (lua_State *L, int narg, off64_t def);
LUALIB_API int32_t (luaL_checkint32_t) (lua_State *L, int narg);
LUALIB_API uint32_t (luaL_checkuint32_t) (lua_State *L, int narg);
LUALIB_API off64_t (luaL_checkoff64_t) (lua_State *L, int narg);

LUALIB_API void (luaL_checkstack) (lua_State *L, int sz, const char *msg);
LUALIB_API void (luaL_checktype) (lua_State *L, int narg, int t);
LUALIB_API void (luaL_checkany) (lua_State *L, int narg);

LUALIB_API int   (luaL_newmetatable) (lua_State *L, const char *tname);
LUALIB_API void *(luaL_checkudata) (lua_State *L, int ud, const char *tname);
LUALIB_API void *(luaL_getudata) (lua_State *L, int ud, const char *tname, int *result);  /* 0.12.2 */
LUALIB_API int   (luaL_isudata) (lua_State *L, int ud, const char *tname);  /* 2.14.6 */

LUALIB_API void (luaL_where) (lua_State *L, int lvl);
LUALIB_API int (luaL_error) (lua_State *L, const char *fmt, ...);

LUALIB_API int (luaL_checkoption) (lua_State *L, int narg, const char *def,
                                   const char *const lst[]);
LUALIB_API int (luaL_ref) (lua_State *L, int t);
LUALIB_API void (luaL_unref) (lua_State *L, int t, int ref);

LUALIB_API int (luaL_loadfile) (lua_State *L, const char *filename);
LUALIB_API int (luaL_loadbuffer) (lua_State *L, const char *buff, size_t sz,
                                  const char *name);
LUALIB_API int (luaL_loadstring) (lua_State *L, const char *s);

LUALIB_API lua_State *(luaL_newstate) (void);


LUALIB_API const char *(luaL_gsub) (lua_State *L, const char *s, const char *p,
                                                  const char *r);

LUALIB_API const char *(luaL_findtable) (lua_State *L, int idx,
                                         const char *fname, int szhint);

LUALIB_API void tag_error (lua_State *L, int narg, int tag);


LUALIB_API void (agnL_setLibname) (lua_State *L, int warning, int debug);
LUALIB_API void (agnL_initialise) (lua_State *L, int skipinis, int debug, int skipmainlib);
LUALIB_API const char *(luaL_pushnexttemplate) (lua_State *L, const char *path);

LUALIB_API void (agnL_printnonstruct) (lua_State *L, int i);
LUALIB_API int  (agnL_gettablefield) (lua_State *L, const char *table, const char *field, const char *procname, int issueerror);  /* 1.6.4 */

LUALIB_API const char *(agnL_optstring) (lua_State *L, int narg, const char *def);                      /* 2.9.0 */
LUALIB_API LUA_UINT32 (agnL_optuint32_t) (lua_State *L, int narg, uint32_t def);
LUALIB_API int (agnL_checkoption) (lua_State *L, int narg, const char *def, const char *const lst[], int ignorecase);  /* 2.9.0/2.16.1 */

LUALIB_API int (agnL_gettop) (lua_State *L, int errnargs, const char *message, const char *procname);  /* 2.9.8 */

LUALIB_API int (agnL_createpairofnumbers) (lua_State *L, lua_Number l, lua_Number r);  /* 2.9.8 */

LUALIB_API lua_Number (agnL_fncall) (lua_State *L, int idx, lua_Number x, int optstart, int optstop, int nresults);  /* 2.10.4 */
LUA_API lua_Number (agnL_fneps) (lua_State *L, int fidx, lua_Number x, int n, int p, int q, lua_Number *origh, lua_Number *abserr);  /* 2.14.3 */
LUALIB_API int  (agnL_islinalgvector) (lua_State *L, int idx, size_t *dim);  /* 2.3.0 RC 4 */
LUALIB_API int  (agnL_fillarray) (lua_State *L, int idx, int type, lua_Number *a, size_t *ii, int left, int right, int reverse);  /* 2.10.4 */
LUALIB_API void (agnL_onexit) (lua_State *L);  /* 2.7.0 */

LUALIB_API void (agnL_readlines) (lua_State *L, FILE *f, const char *procname, int ispipe);  /* 2.16.10 */
LUALIB_API int  (agnL_pexecute) (lua_State *L, const char *str, const char *procname);  /* 2.16.10 */

/*
** ===============================================================
** some useful macros
** ===============================================================
*/

#define luaL_argcheck(L, cond,numarg,extramsg)	\
		((void)((cond) || luaL_argerror(L, (numarg), (extramsg))))
#define luaL_typecheck(L, cond,numarg,extramsg,expectedtype)	\
		((void)((cond) || luaL_typeerror(L, (numarg), (extramsg), (expectedtype))))
#define luaL_checkstring(L,n)	(luaL_checklstring(L, (n), NULL))
#define luaL_optstring(L,n,d)	(luaL_optlstring(L, (n), (d), NULL))
#define luaL_checkint(L,n)	((int)luaL_checkinteger(L, (n)))
#define luaL_optint(L,n,d)	((int)luaL_optinteger(L, (n), (d)))
#define luaL_checklong(L,n)	((long)luaL_checkinteger(L, (n)))
#define agnL_checkint(L,n)	((int)agnL_checkinteger(L, (n)))  /* Agena 1.4.3/1.5.0 */
#define luaL_optlong(L,n,d)	((long)luaL_optinteger(L, (n), (d)))

#define luaL_typename(L,i)	lua_typename(L, lua_type(L,(i)))

#define luaL_dofile(L, fn) \
	(luaL_loadfile(L, fn) || lua_pcall(L, 0, LUA_MULTRET, 0))

#define luaL_dostring(L, s) \
	(luaL_loadstring(L, s) || lua_pcall(L, 0, LUA_MULTRET, 0))

#define luaL_getmetatable(L,n)	(lua_getfield(L, LUA_REGISTRYINDEX, (n)))

#define luaL_opt(L,f,n,d)	(lua_isnoneornil(L,(n)) ? (d) : f(L,(n)))

#define _agnc_mul(x,y) { \
  agn_Complex r; \
  lua_Number a, b, c, d; \
  a = creal(x); b = cimag(x); c = creal(y); d = cimag(y); \
  __real__ r = a*c - b*d; \
  __imag__ r = a*d + b*c; \
  (r); \
}

/* 2.16.9 , the macro slower than a C function; see: https://gcc.gnu.org/onlinedocs/gcc/Statement-Exprs.html:
   we use an assignment to prevent (narg) computed twice.
#define magnL_optuint32_t(L, narg, def) \
  ( { int __nargoptuint32 = narg; \
  (uint32_t)( lua_isnoneornil(L, (__nargoptuint32)) ? (def) : agn_checknonnegative(L, (__nargoptuint32)) ); } ) */

/*
** {======================================================
** Generic Buffer manipulation
** =======================================================
*/

typedef struct luaL_Buffer {
  char *p;	/* current position in buffer */
  int lvl;  /* number of strings in the stack (level) */
  lua_State *L;
  char buffer[LUAL_BUFFERSIZE];
  /* taken from Lua 5.3.5: */
  char *b;  /* buffer address */
  size_t size;  /* buffer size */
  size_t n;  /* number of characters in buffer */
  /* char initb[LUAL_BUFFERSIZE]; */ /* initial buffer */
} luaL_Buffer;

#define luaL_addchar(B,c) \
  ((void)((B)->p < ((B)->buffer+LUAL_BUFFERSIZE) || luaL_prepbuffer(B)), \
   (*(B)->p++ = (char)(c)))

#define luaL_addchar53(B,c) \
  ((void)((B)->n < (B)->size || luaL_prepbuffsize((B), 1)), \
   ((B)->b[(B)->n++] = (c)))

#define luaL_addsize53(B,s)	((B)->n += (s))

/* compatibility only */
#define luaL_putchar(B,c)	luaL_addchar(B,c)

#define luaL_addsize(B,n)	((B)->p += (n))

LUALIB_API void (luaL_buffinit) (lua_State *L, luaL_Buffer *B);
LUALIB_API char *(luaL_prepbuffer) (luaL_Buffer *B);
LUALIB_API void (luaL_addlstring) (luaL_Buffer *B, const char *s, size_t l);
LUALIB_API void (luaL_addstring) (luaL_Buffer *B, const char *s);
LUALIB_API void (luaL_addvalue) (luaL_Buffer *B);
LUALIB_API void (luaL_pushresult) (luaL_Buffer *B);
LUALIB_API void (luaL_clearbuffer) (luaL_Buffer *B);

LUALIB_API int emptybuffer (luaL_Buffer *B);

#define bufflen(B)	((B)->p - (B)->buffer)

LUALIB_API char *luaL_prepbuffsize (luaL_Buffer *B, size_t sz);
LUALIB_API void luaL_addlstring53 (luaL_Buffer *B, const char *s, size_t l);
LUALIB_API void luaL_addstring53 (luaL_Buffer *B, const char *s);
LUALIB_API void luaL_pushresult53 (luaL_Buffer *B);

/* }====================================================== */


/* compatibility with ref system */

/* pre-defined references */
#define LUA_NOREF       (-2)
#define LUA_REFNIL      (-1)

#define lua_ref(L,lock) ((lock) ? luaL_ref(L, LUA_REGISTRYINDEX) : \
      (lua_pushstring(L, "unlocked references are obsolete"), lua_error(L), 0))

#define lua_unref(L,ref)        luaL_unref(L, LUA_REGISTRYINDEX, (ref))

#define lua_getref(L,ref)       lua_rawgeti(L, LUA_REGISTRYINDEX, (ref))


#define luaL_reg	luaL_Reg

#endif

